import re
import shlex
import click

import keyring
from pygit2 import Repository

from .jira import JIRA, JIRAError
from .harvest import HarvestPlatform, HarvestAPI, HarvestError
from .utils import cached_property, taskstatus


class MappedProjectID:
    def __init__(self, project_ids):
        self._project_ids = project_ids

    @classmethod
    def fromstring(cls, string):
        project_ids = string.split(',')
        project_ids = (p.split(':') for p in project_ids)
        project_ids = ((None, p) if len(p) == 1 else p for p in project_ids)
        project_ids = dict(project_ids)
        return cls(project_ids)

    def __call__(self, issue):
        try:
            return self._project_ids[str(issue.fields.issuetype)]
        except KeyError:
            pass

        try:
            return self._project_ids[None]
        except KeyError:
            pass

        raise ValueError('Could not find a project ID for issue type "{}"'
                         .format(issue.fields.issuetype))


class NullIntegrationHelper:
    def register(self, *args, **kwargs):
        pass

    def close(self):
        pass


class WarnIntegrationHelper(NullIntegrationHelper):
    def __init__(self):
        self._shown = False

    def register(self, *args, **kwargs):
        if not self._shown:
            self._shown = True
            click.secho('')
            click.secho('  Lancet executable called directly', fg='yellow')
            click.secho('  ---------------------------------', fg='yellow')
            click.secho('')
            click.secho('  Setup the shell integration to enjoy some of the')
            click.secho('  super powers we built right into lancet.')
            click.secho('')
            click.secho('  This basically means to add the following snippet')
            click.secho('  to your shell initialization file:')
            click.secho('')
            click.secho('    lancet --setup-helper | source /dev/stdin')
            click.secho('')
            click.secho('  See {} for addtional details.'.format(
                click.style('https://lancet.rtd.org', fg='green')))
            click.secho('')


class ShellIntegrationHelper(NullIntegrationHelper):
    def __init__(self, filename):
        self.filename = filename
        self.fh = open(filename, 'w')

    def register(self, *args, raw=False):
        cmd = args[0] if raw else ' '.join(shlex.quote(a) for a in args)
        self.fh.write(cmd)
        self.fh.write('\n')

    def close(self):
        self.fh.close()


class Lancet:
    def __init__(self, config, integration_helper):
        self.config = config
        self.integration_helper = integration_helper

    def defer_to_shell(self, *args, **kwargs):
        return self.integration_helper.register(*args, **kwargs)

    @cached_property
    def repo(self):
        # Can be cleared like this: self.__class__.repo.fget.cache_clear()
        return Repository('./.git')

    def get_credentials(self, service, checker=None):
        url = self.config.get(service, 'url')
        username = self.config.get(service, 'username')
        key = 'lancet+{}'.format(url)
        if username:
            password = keyring.get_password(key, username)
            if password:
                return url, username, password

        with taskstatus.suspend():
            while True:
                click.echo(
                    'Please provide your authentication information for {}'
                    .format(url)
                )
                if not username:
                    username = click.prompt('Username')
                else:
                    click.echo('Username: {}'.format(username))
                password = click.prompt('Password', hide_input=True)

                if checker:
                    with taskstatus('Checking provided credentials') as ts:
                        if not checker(url, username, password):
                            ts.fail('Login failed')
                            username, password = None, None
                            continue
                        else:
                            ts.ok('Correctly authenticated to {}', url)

                keyring.set_password(key, username, password)
                return url, username, password

    def get_issue(self, key=None):
        # TODO: Move this method to the JIRA class

        if key is None:
            # TODO: This should be factored out of here
            match = re.search(r'feature/([A-Z]{2,}-[0-9]+)',
                              self.repo.head.name)
            if match is None:
                raise Exception('Unable to find current issue.')
            key = match.group(1)
        elif key.isdigit():
            project_key = self.config.get('tracker', 'default_project')
            if project_key:
                key = '{}-{}'.format(project_key, key)
        return self.tracker.issue(key)

    @cached_property
    def tracker(self):
        def checker(url, username, password):
            try:
                JIRA(server=url, basic_auth=(username, password))
            except JIRAError:
                return False
            else:
                return True

        url, username, password = self.get_credentials('tracker', checker)
        tracker = JIRA(server=url, basic_auth=(username, password))
        self.call_on_close(tracker.close)
        return tracker

    @cached_property
    def timer(self):
        def checker(url, username, password):
            api = HarvestAPI(url, (username, password))
            try:
                api.whoami()
            except HarvestError:
                return False
            else:
                return True

        url, username, password = self.get_credentials('harvest', checker)
        task_id = self.config.get('harvest', 'task_id')
        project_id_getter = MappedProjectID.fromstring(
            self.config.get('harvest', 'project_id'))
        timer = HarvestPlatform(server=url,
                                basic_auth=(username, password),
                                project_id_getter=project_id_getter,
                                task_id=task_id)
        self.call_on_close(timer.close)
        return timer



import functools

from html.parser import HTMLParser
from urllib.parse import urljoin, urlencode
import requests


class MetaRetriever(HTMLParser):
    def __init__(self, name, *args, **kwargs):
        self.meta_name = name
        self.meta_content = ''
        kwargs['convert_charrefs'] = True
        super().__init__(*args, **kwargs)

    def handle_starttag(self, tag, attrs):
        attrs = dict(attrs)
        if tag == 'meta' and attrs.get('name', None) == self.meta_name:
            self.meta_content = attrs.get('content', '')


def get_meta_content(html, name):
    retriever = MetaRetriever(name)
    retriever.feed(html)
    return retriever.meta_content


def requires_login(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        self._login()
        return func(self, *args, **kwargs)
    return wrapper


class HarvestError(Exception):
    pass


class HarvestAPI:
    def __init__(self, server, basic_auth):
        self.server = server
        self.auth = basic_auth
        self._session = requests.Session()
        self._session.auth = basic_auth
        self._session.headers = {
            'accept': 'application/json'
        }

    def _get(self, url):
        r = self._session.get(urljoin(self.server, url))
        payload = r.json()
        if r.status_code != 200:
            raise HarvestError(payload['message'])
        return payload

    def _post(self, url, data):
        return self._session.post(urljoin(self.server, url)).json()

    def toggle(self, id):
        return self._get('daily/timer/{}'.format(id))

    def pause(self, id=None):
        for entry in self.daily():
            if 'timer_started_at' in entry:
                if id is None or id == entry['id']:
                    self.toggle(entry['id'])
                break
        else:
            # No running timers
            pass

    def whoami(self):
        return self._get('account/who_am_i')

    def projects(self):
        return self._get('daily')['projects']

    def daily(self):
        return self._get('daily')['day_entries']

    def close(self):
        self._session.close()


class HarvestPlatform(HarvestAPI):
    platform_url = 'https://platform.harvestapp.com'

    def __init__(self, server, basic_auth, project_id_getter, task_id):
        self._web_session = requests.Session()
        self.task_id = task_id
        self.get_project_id = project_id_getter
        self._csrf_token = None
        super().__init__(server, basic_auth)

    def _get_csrf_token(self):
        if not self._csrf_token:
            r = self._web_session.get(urljoin(self.server, '/account/login'))
            self._csrf_token = get_meta_content(r.text, 'csrf-token')
        return self._csrf_token

    def _login(self):
        # TODO: Cache the harvest session across commands invocations
        if '_harvest_sess' in self._web_session.cookies:
            return
        data = {
            'utf8': '✓',
            'authenticity_token': self._get_csrf_token(),
            'user[email]': self.auth[0],
            'user[password]': self.auth[1],
        }
        self._web_session.post(
            urljoin(self.server, '/account/create_session'),
            data=data,
        )
        self._csrf_token = None

    @requires_login
    def start(self, issue, resume=True):
        project = issue.fields.project

        if resume:
            for entry in reversed(self.daily()):
                ext = entry.get('external_ref', None)
                if not ext:
                    continue
                if ext['group_id'] != str(project.id):
                    continue
                if ext['id'] != str(issue.id):
                    continue
                if 'timer_started_at' not in entry:
                    self.toggle(entry['id'])
                return

        name = '{} - {}'.format(issue.key, issue.fields.summary)

        qs = urlencode({
            'app_name': 'JIRA',
            'base_url': issue.permalink(),
            'external_account_id': 1,  # ID of the JIRA instance
            'external_group_id': project.id,
            'external_group_name': project.name,
            'external_item_id': issue.id,
            'external_item_name': name,
            'service': 'divio-ch.atlassian.net',
        })
        data = {
            'utf8': '✓',
            'project_id': self.get_project_id(issue),
            'task_id': self.task_id,
            'notes': name,
            'hours': '',
            'button': '',
        }
        headers = {
            'x-csrf-token': self._get_csrf_token(),
            'x-requested-with': 'XMLHttpRequest',
        }
        r = self._web_session.post(
            urljoin(self.platform_url, '/platform/timer?{}'.format(qs)),
            data=data,
            headers=headers,
        )
        assert r.status_code == 200
        assert 'message' in r.json()

    def close(self):
        self._web_session.close()
        super(HarvestPlatform, self).close()



import sys

import pygit2
import click
from slugify import slugify

from .utils import taskstatus


class SlugBranchGetter(object):
    prefix = 'feature/'

    def __init__(self, base_branch, remote_credentials, remote_name='origin'):
        self.base_branch = base_branch
        self.remote_name = remote_name
        self.remote_credentials = remote_credentials

    def get_base_branch(self, repo):
        return repo.lookup_branch(
            '{}/{}'.format(self.remote_name, self.base_branch),
            pygit2.GIT_BRANCH_REMOTE
        )

    def get_branch_name(self, prefix, issue):
        discriminator = '{}{}_'.format(prefix, issue.key)
        slug = slugify(issue.fields.summary[:30])
        full_name = '{}{}_{}'.format(self.prefix, issue.key, slug)
        return discriminator, full_name

    def get_branch(self, repo, issue, from_remote=False):
        if from_remote:
            branch_type = pygit2.GIT_BRANCH_REMOTE
            prefix = '{}/{}'.format(self.remote_name, self.prefix)
        else:
            branch_type = pygit2.GIT_BRANCH_LOCAL
            prefix = self.prefix

        discriminator, full_name = self.get_branch_name(prefix, issue)

        branches = [b for b in repo.listall_branches(branch_type)
                    if b.startswith(discriminator)]

        if len(branches) > 1:
            click.secho('Multiple matching branches found!',
                        fg='red', bold=True)
            click.echo()
            click.echo('The prefix {} matched the following branches:'
                       .format(discriminator))
            click.echo()
            for b in branches:
                click.echo(' {} {}'.format(click.style('*', fg='red'), b))
            click.echo()
            click.echo('Please remove all but one in order to continue.')
            sys.exit(1)
        elif branches:
            branch = repo.lookup_branch(branches[0], branch_type)
            if from_remote:
                branch = repo.create_branch(full_name, branch.get_object())
            elif branch.branch_name != full_name:
                # We only need to rename if we're working on an already
                # existing local branch
                branch.rename(full_name)
                branch = repo.lookup_branch(full_name)
            return branch

    def __call__(self, repo, issue):
        branch = self.get_branch(repo, issue)

        if not branch:
            # No local branches found, fetch remote
            with taskstatus('Fetching from "{}"', self.remote_name) as ts:
                for remote in repo.remotes:
                    if remote.name == self.remote_name:
                        break
                else:
                    ts.abort('Remote "{}" not found', self.remote_name)

                remote.credentials = self.remote_credentials
                remote.fetch()
                ts.ok('Fetched latest changes from "{}"', self.remote_name)

            # Check remote branches
            with taskstatus('Creating working branch') as ts:
                branch = self.get_branch(repo, issue, from_remote=True)

                # Create off origin base branch
                if branch:
                    ts.ok('Created new working branch based on existing '
                          'remote branch')
                else:
                    _, full_name = self.get_branch_name(self.prefix, issue)

                    base = self.get_base_branch(repo)
                    if not base:
                        ts.abort('Base branch "{}" not found on remote "{}"',
                                 self.base_branch, self.remote_name)
                    branch = repo.create_branch(full_name, base.get_object())
                    ts.ok('Created new working branch')

        return branch
